@@ -1,8 +1,9 @@
 from __future__ import with_statement
-from itertools import starmap
+from itertools import chain, starmap
 import datetime
-import time
+import sys
 import warnings
+import time as mod_time
 from redis._compat import (b, izip, imap, iteritems, dictkeys, dictvalues,
                            basestring, long, nativestr, urlparse)
 from redis.connection import ConnectionPool, UnixDomainSocketConnection
@@ -12,6 +13,7 @@ from redis.exceptions import (
     RedisError,
     ResponseError,
     WatchError,
+    NoScriptError
 )
 
 SYM_EMPTY = b('')
@@ -139,14 +141,33 @@ def float_or_none(response):
     return float(response)
 
 
+def parse_client(response, **options):
+    parse = options['parse']
+    if parse == 'LIST':
+        clients = []
+        for c in nativestr(response).splitlines():
+            clients.append(dict([pair.split('=') for pair in c.split(' ')]))
+        return clients
+    elif parse == 'KILL':
+        return bool(response)
+
+
 def parse_config(response, **options):
-    # this is stupid, but don't have a better option right now
     if options['parse'] == 'GET':
         response = [nativestr(i) if i is not None else None for i in response]
         return response and pairs_to_dict(response) or {}
     return nativestr(response) == 'OK'
 
 
+def parse_script(response, **options):
+    parse = options['parse']
+    if parse in ('FLUSH', 'KILL'):
+        return response == 'OK'
+    if parse == 'EXISTS':
+        return list(imap(bool, response))
+    return response
+
+
 class StrictRedis(object):
     """
     Implementation of the Redis protocol.
@@ -164,11 +185,12 @@ class StrictRedis(object):
             bool
         ),
         string_keys_to_dict(
-            'DECRBY GETBIT HLEN INCRBY LINSERT LLEN LPUSHX RPUSHX SADD SCARD '
-            'SDIFFSTORE SETBIT SETRANGE SINTERSTORE STRLEN SUNIONSTORE ZADD '
-            'ZCARD ZREMRANGEBYRANK ZREMRANGEBYSCORE',
+            'BITCOUNT DECRBY GETBIT HLEN INCRBY LINSERT LLEN LPUSHX RPUSHX '
+            'SADD SCARD SDIFFSTORE SETBIT SETRANGE SINTERSTORE STRLEN '
+            'SUNIONSTORE ZADD ZCARD ZREMRANGEBYRANK ZREMRANGEBYSCORE',
             int
         ),
+        string_keys_to_dict('INCRBYFLOAT HINCRBYFLOAT', float),
         string_keys_to_dict(
             # these return OK, or int if redis-server is >=1.3.4
             'LPUSH RPUSH',
@@ -196,6 +218,7 @@ class StrictRedis(object):
             ),
             'BGSAVE': lambda r: r == 'Background saving started',
             'BRPOPLPUSH': lambda r: r and r or None,
+            'CLIENT': parse_client,
             'CONFIG': parse_config,
             'DEBUG': parse_debug_object,
             'HGETALL': lambda r: r and pairs_to_dict(r) or {},
@@ -204,6 +227,8 @@ class StrictRedis(object):
             'OBJECT': parse_object,
             'PING': lambda r: nativestr(r) == 'PONG',
             'RANDOMKEY': lambda r: r and r or None,
+            'SCRIPT': parse_script,
+            'TIME': lambda x: (int(x[0]), int(x[1]))
         }
     )
 
@@ -359,6 +384,14 @@ class StrictRedis(object):
         """
         return self.execute_command('BGSAVE')
 
+    def client_kill(self, address):
+        "Disconnects the client at ``address`` (ip:port)"
+        return self.execute_command('CLIENT', 'KILL', address, parse='KILL')
+
+    def client_list(self):
+        "Returns a list of currently connected clients"
+        return self.execute_command('CLIENT', 'LIST', parse='LIST')
+
     def config_get(self, pattern="*"):
         "Return a dictionary of configuration based on the ``pattern``"
         return self.execute_command('CONFIG', 'GET', pattern, parse='GET')
@@ -371,6 +404,13 @@ class StrictRedis(object):
         "Returns the number of keys in the current database"
         return self.execute_command('DBSIZE')
 
+    def time(self):
+        """
+        Returns the server time as a 2-item tuple of ints:
+        (seconds since epoch, microseconds into this second).
+        """
+        return self.execute_command('TIME')
+
     def debug_object(self, key):
         "Returns version specific metainformation about a give key"
         return self.execute_command('DEBUG', 'OBJECT', key)
@@ -453,6 +493,26 @@ class StrictRedis(object):
         """
         return self.execute_command('GETRANGE', key, start, end)
 
+    def bitcount(self, key, start=None, end=None):
+        """
+        Returns the count of set bits in the value of ``key``.  Optional
+        ``start`` and ``end`` paramaters indicate which bytes to consider
+        """
+        params = [key]
+        if start and end:
+            params.append(start)
+            params.append(end)
+        elif (start and not end) or (end and not start):
+            raise RedisError("Both start and end must be specified")
+        return self.execute_command('BITCOUNT', *params)
+
+    def bitop(self, operation, dest, *keys):
+        """
+        Perform a bitwise operation using ``operation`` between ``keys`` and
+        store the result in ``dest``.
+        """
+        return self.execute_command('BITOP', operation, dest, *keys)
+
     def decr(self, name, amount=1):
         """
         Decrements the value of ``key`` by ``amount``.  If no key exists,
@@ -471,7 +531,7 @@ class StrictRedis(object):
         can be represented by an integer or a Python timedelta object.
         """
         if isinstance(time, datetime.timedelta):
-            time = int(time.total_seconds())
+            time = time.seconds + time.days * 24 * 3600
         return self.execute_command('EXPIRE', name, time)
 
     def expireat(self, name, when):
@@ -480,7 +540,7 @@ class StrictRedis(object):
         as an integer indicating unix time or a Python datetime object.
         """
         if isinstance(when, datetime.datetime):
-            when = int(time.mktime(when.timetuple()))
+            when = int(mod_time.mktime(when.timetuple()))
         return self.execute_command('EXPIREAT', name, when)
 
     def get(self, name):
@@ -517,6 +577,13 @@ class StrictRedis(object):
         """
         return self.execute_command('INCRBY', name, amount)
 
+    def incrbyfloat(self, name, amount=1.0):
+        """
+        Increments the value at key ``name`` by floating ``amount``.
+        If no key exists, the value will be initialized as ``amount``
+        """
+        return self.execute_command('INCRBYFLOAT', name, amount)
+
     def keys(self, pattern='*'):
         "Returns a list of keys matching ``pattern``"
         return self.execute_command('KEYS', pattern)
@@ -553,6 +620,32 @@ class StrictRedis(object):
         "Removes an expiration on ``name``"
         return self.execute_command('PERSIST', name)
 
+    def pexpire(self, name, time):
+        """
+        Set an expire flag on key ``name`` for ``time`` milliseconds.
+        ``time`` can be represented by an integer or a Python timedelta
+        object.
+        """
+        if isinstance(time, datetime.timedelta):
+            ms = int(time.microseconds / 1000)
+            time = time.seconds + time.days * 24 * 3600 * 1000 + ms
+        return self.execute_command('PEXPIRE', name, time)
+
+    def pexpireat(self, name, when):
+        """
+        Set an expire flag on key ``name``. ``when`` can be represented
+        as an integer representing unix time in milliseconds (unix time * 1000)
+        or a Python datetime object.
+        """
+        if isinstance(when, datetime.datetime):
+            ms = int(when.microsecond / 1000)
+            when = int(mod_time.mktime(when.timetuple())) * 1000 + ms
+        return self.execute_command('PEXPIREAT', name, when)
+
+    def pttl(self, name):
+        "Returns the number of milliseconds until the key ``name`` will expire"
+        return self.execute_command('PTTL', name)
+
     def randomkey(self):
         "Returns the name of a random key"
         return self.execute_command('RANDOMKEY')
@@ -587,7 +680,7 @@ class StrictRedis(object):
         timedelta object.
         """
         if isinstance(time, datetime.timedelta):
-            time = int(time.total_seconds())
+            time = time.seconds + time.days * 24 * 3600
         return self.execute_command('SETEX', name, time, value)
 
     def setnx(self, name, value):
@@ -888,9 +981,16 @@ class StrictRedis(object):
         "Remove and return a random member of set ``name``"
         return self.execute_command('SPOP', name)
 
-    def srandmember(self, name):
-        "Return a random member of set ``name``"
-        return self.execute_command('SRANDMEMBER', name)
+    def srandmember(self, name, number=None):
+        """
+        If ``number`` is None, returns a random member of set ``name``.
+
+        If ``number`` is supplied, returns a list of ``number`` random
+        memebers of set ``name``. Note this is only available when running
+        Redis 2.6+.
+        """
+        args = number and [number] or []
+        return self.execute_command('SRANDMEMBER', name, *args)
 
     def srem(self, name, *values):
         "Remove ``values`` from set ``name``"
@@ -1130,6 +1230,12 @@ class StrictRedis(object):
         "Increment the value of ``key`` in hash ``name`` by ``amount``"
         return self.execute_command('HINCRBY', name, key, amount)
 
+    def hincrbyfloat(self, name, key, amount=1.0):
+        """
+        Increment the value of ``key`` in hash ``name`` by floating ``amount``
+        """
+        return self.execute_command('HINCRBYFLOAT', name, key, amount)
+
     def hkeys(self, name):
         "Return the list of keys within hash ``name``"
         return self.execute_command('HKEYS', name)
@@ -1180,6 +1286,61 @@ class StrictRedis(object):
         """
         return self.execute_command('PUBLISH', channel, message)
 
+    def eval(self, script, numkeys, *keys_and_args):
+        """
+        Execute the LUA ``script``, specifying the ``numkeys`` the script
+        will touch and the key names and argument values in ``keys_and_args``.
+        Returns the result of the script.
+
+        In practice, use the object returned by ``register_script``. This
+        function exists purely for Redis API completion.
+        """
+        return self.execute_command('EVAL', script, numkeys, *keys_and_args)
+
+    def evalsha(self, sha, numkeys, *keys_and_args):
+        """
+        Use the ``sha`` to execute a LUA script already registered via EVAL
+        or SCRIPT LOAD. Specify the ``numkeys`` the script will touch and the
+        key names and argument values in ``keys_and_args``. Returns the result
+        of the script.
+
+        In practice, use the object returned by ``register_script``. This
+        function exists purely for Redis API completion.
+        """
+        return self.execute_command('EVALSHA', sha, numkeys, *keys_and_args)
+
+    def script_exists(self, *args):
+        """
+        Check if a script exists in the script cache by specifying the SHAs of
+        each script as ``args``. Returns a list of boolean values indicating if
+        if each already script exists in the cache.
+        """
+        options = {'parse': 'EXISTS'}
+        return self.execute_command('SCRIPT', 'EXISTS', *args, **options)
+
+    def script_flush(self):
+        "Flush all scripts from the script cache"
+        options = {'parse': 'FLUSH'}
+        return self.execute_command('SCRIPT', 'FLUSH', **options)
+
+    def script_kill(self):
+        "Kill the currently executing LUA script"
+        options = {'parse': 'KILL'}
+        return self.execute_command('SCRIPT', 'KILL', **options)
+
+    def script_load(self, script):
+        "Load a LUA ``script`` into the script cache. Returns the SHA."
+        options = {'parse': 'LOAD'}
+        return self.execute_command('SCRIPT', 'LOAD', script, **options)
+
+    def register_script(self, script):
+        """
+        Register a LUA ``script`` specifying the ``keys`` it will touch.
+        Returns a Script object that is callable and hides the complexity of
+        deal with scripts, keys, and shas. This is the preferred way to work
+        with LUA scripts.
+        """
+        return Script(self, script)
 
 class Redis(StrictRedis):
     """
@@ -1193,6 +1354,7 @@ class Redis(StrictRedis):
         StrictRedis.RESPONSE_CALLBACKS,
         {
             'TTL': lambda r: r != -1 and r or None,
+            'PTTL': lambda r: r != -1 and r or None,
         }
     )
 
@@ -1217,7 +1379,7 @@ class Redis(StrictRedis):
         timedelta object.
         """
         if isinstance(time, datetime.timedelta):
-            time = int(time.total_seconds())
+            time = time.seconds + time.days * 24 * 3600
         return self.execute_command('SETEX', name, time, value)
 
     def lrem(self, name, value, num=0):
@@ -1454,6 +1616,7 @@ class BasePipeline(object):
 
     def reset(self):
         self.command_stack = []
+        self.scripts = set()
         # make sure to reset the connection state in the event that we were
         # watching something
         if self.watching and self.connection:
@@ -1534,27 +1697,40 @@ class BasePipeline(object):
         self.command_stack.append((args, options))
         return self
 
-    def _execute_transaction(self, connection, commands):
+    def _execute_transaction(self, connection, commands, raise_on_error):
+        cmds = chain([(('MULTI', ), {})], commands, [(('EXEC', ), {})])
         all_cmds = SYM_EMPTY.join(
             starmap(connection.pack_command,
-                    [args for args, options in commands]))
+                    [args for args, options in cmds]))
         connection.send_packed_command(all_cmds)
-        # we don't care about the multi/exec any longer
-        commands = commands[1:-1]
-        # parse off the response for MULTI and all commands prior to EXEC.
-        # the only data we care about is the response the EXEC
-        # which is the last command
-        for i in range(len(commands) + 1):
-            self.parse_response(connection, '_')
+        # parse off the response for MULTI
+        self.parse_response(connection, '_')
+        # and all the other commands
+        errors = []
+        for i, _ in enumerate(commands):
+            try:
+                self.parse_response(connection, '_')
+            except ResponseError:
+                errors.append((i, sys.exc_info()[1]))
+
         # parse the EXEC.
         response = self.parse_response(connection, '_')
 
         if response is None:
             raise WatchError("Watched variable changed.")
 
+        # put any parse errors into the response
+        for i, e in errors:
+            response.insert(i, e)
+
         if len(response) != len(commands):
             raise ResponseError("Wrong number of response items from "
                                 "pipeline execution")
+
+        # find any errors in the response and raise if necessary
+        if raise_on_error:
+            self.raise_first_error(response)
+
         # We have to run response callbacks manually
         data = []
         for r, cmd in izip(response, commands):
@@ -1566,14 +1742,22 @@ class BasePipeline(object):
             data.append(r)
         return data
 
-    def _execute_pipeline(self, connection, commands):
+    def _execute_pipeline(self, connection, commands, raise_on_error):
         # build up all commands into a single request to increase network perf
         all_cmds = SYM_EMPTY.join(
             starmap(connection.pack_command,
                     [args for args, options in commands]))
         connection.send_packed_command(all_cmds)
-        return [self.parse_response(connection, args[0], **options)
-                for args, options in commands]
+        response = [self.parse_response(connection, args[0], **options)
+                    for args, options in commands]
+        if raise_on_error:
+            self.raise_first_error(response)
+        return response
+
+    def raise_first_error(self, response):
+        for r in response:
+            if isinstance(r, ResponseError):
+                raise r
 
     def parse_response(self, connection, command_name, **options):
         result = StrictRedis.parse_response(
@@ -1584,11 +1768,23 @@ class BasePipeline(object):
             self.watching = True
         return result
 
-    def execute(self):
+    def load_scripts(self):
+        # make sure all scripts that are about to be run on this pipeline exist
+        scripts = list(self.scripts)
+        immediate = self.immediate_execute_command
+        shas = [s.sha for s in scripts]
+        exists = immediate('SCRIPT', 'EXISTS', *shas, **{'parse': 'EXISTS'})
+        if not all(exists):
+            for s, exist in izip(scripts, exists):
+                if not exist:
+                    immediate('SCRIPT', 'LOAD', s.script, **{'parse': 'LOAD'})
+
+    def execute(self, raise_on_error=True):
         "Execute all the commands in the current pipeline"
+        if self.scripts:
+            self.load_scripts()
         stack = self.command_stack
         if self.transaction or self.explicit_transaction:
-            stack = [(('MULTI', ), {})] + stack + [(('EXEC', ), {})]
             execute = self._execute_transaction
         else:
             execute = self._execute_pipeline
@@ -1602,7 +1798,7 @@ class BasePipeline(object):
             self.connection = conn
 
         try:
-            return execute(conn, stack)
+            return execute(conn, stack, raise_on_error)
         except ConnectionError:
             conn.disconnect()
             # if we were watching a variable, the watch is no longer valid
@@ -1615,24 +1811,24 @@ class BasePipeline(object):
                                  "one or more keys")
             # otherwise, it's safe to retry since the transaction isn't
             # predicated on any state
-            return execute(conn, stack)
+            return execute(conn, stack, raise_on_error)
         finally:
             self.reset()
 
     def watch(self, *names):
-        """
-        Watches the values at keys ``names``
-        """
+        "Watches the values at keys ``names``"
         if self.explicit_transaction:
             raise RedisError('Cannot issue a WATCH after a MULTI')
         return self.execute_command('WATCH', *names)
 
     def unwatch(self):
-        """
-        Unwatches all previously specified keys
-        """
+        "Unwatches all previously specified keys"
         return self.watching and self.execute_command('UNWATCH') or True
 
+    def script_load_for_pipeline(self, script):
+        "Make sure scripts are loaded prior to pipeline execution"
+        self.scripts.add(script)
+
 
 class StrictPipeline(BasePipeline, StrictRedis):
     "Pipeline for the StrictRedis class"
@@ -1644,6 +1840,31 @@ class Pipeline(BasePipeline, Redis):
     pass
 
 
+class Script(object):
+    "An executable LUA script object returned by ``register_script``"
+
+    def __init__(self, registered_client, script):
+        self.registered_client = registered_client
+        self.script = script
+        self.sha = registered_client.script_load(script)
+
+    def __call__(self, keys=[], args=[], client=None):
+        "Execute the script, passing any required ``args``"
+        client = client or self.registered_client
+        args = tuple(keys) + tuple(args)
+        # make sure the Redis server knows about the script
+        if isinstance(client, BasePipeline):
+            # make sure this script is good to go on pipeline
+            client.script_load_for_pipeline(self)
+        try:
+            return client.evalsha(self.sha, len(keys), *args)
+        except NoScriptError:
+            # Maybe the client is pointed to a differnet server than the client
+            # that created this instance?
+            self.sha = client.script_load(self.script)
+            return client.evalsha(self.sha, len(keys), *args)
+
+
 class LockError(RedisError):
     "Errors thrown from the Lock"
     pass
@@ -1701,7 +1922,7 @@ class Lock(object):
         sleep = self.sleep
         timeout = self.timeout
         while 1:
-            unixtime = int(time.time())
+            unixtime = int(mod_time.time())
             if timeout:
                 timeout_at = unixtime + timeout
             else:
@@ -1722,7 +1943,7 @@ class Lock(object):
                     return True
             if not blocking:
                 return False
-            time.sleep(sleep)
+            mod_time.sleep(sleep)
 
     def release(self):
         "Releases the already acquired lock"
